﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml.Linq;

namespace Goshoom.DynamicsAX
{
    /// <summary>
    /// Parse compilation results exported from Dynamics AX
    /// to a collection of objects.
    /// </summary>
    public class LogParser
    {
        /// <summary>
        /// Parses an .html file with compilation output.
        /// </summary>
        /// <param name="fileName">File name with path</param>
        /// <returns>Collection of faults</returns>
        public IEnumerable<Fault> ParseFile(string fileName)
        {
            if (fileName == null)
            {
                throw new ArgumentNullException("filename");
            }

            return ParseHtml(GetContents(fileName));
        }

        /// <summary>
        /// Parses HTML string.
        /// </summary>
        /// <param name="html">HTML string</param>
        /// <returns>Collection of faults</returns>
        public IEnumerable<Fault> ParseHtml(string html)
        {
            if (html == null)
            {
                throw new ArgumentNullException("html");
            }
            
            return ParseXml(ExtractXml(html));
        }

        /// <summary>
        /// Parses XML string.
        /// </summary>
        /// <param name="xml">XML string</param>
        /// <returns>Collection of faults</returns>
        public IEnumerable<Fault> ParseXml(string xml)
        {
            XDocument loaded = XDocument.Parse(xml);
            XNamespace tableNs = "urn:www.microsoft.com/Formats/Table";

            var faults = from record in loaded.Descendants(tableNs + "Record")
                         let fields = record.Descendants()
                         select new Fault
                         {
                             Column = Int32.Parse(FindValue(fields, "Column")),
                             CompilationTime = DateTime.Parse(FindValue(fields, "createdDateTime")),
                             ElementName = FindValue(fields, "SysUtilElementName"),
                             ElementType = (UtilElementType)Int32.Parse(FindValue(fields, "UtilElementType")),
                             FaultCode = Int32.Parse(FindValue(fields, "CompileErrorCode")),
                             FaultMessage = FindValue(fields, "SysCompileErrorMessage"),
                             FaultString = FindValue(fields, "CompileErrorString"),
                             FaultType = ParseFaultType(fields),
                             Line = Int32.Parse(FindValue(fields, "Line")),
                             MethodName = FindValue(fields, "SysAotMethodName"),
                             PropertyName = FindValue(fields, "SysPropertyName"),
                             TreeNodePath = FindValue(fields, "TreeNodePath")
                         };

            return faults;
        }

        private FaultType ParseFaultType(IEnumerable<XElement> fields)
        {
            CompilerOutputTab tab = (CompilerOutputTab)Int32.Parse(FindValue(fields, "SysCompilerOutputTab"));
            Severity severity = (Severity)Int32.Parse(FindValue(fields, "SysCompilerSeverity"));

            switch (tab)
            {
                case CompilerOutputTab.Error:
                    return severity == Severity.Error ? FaultType.Error : FaultType.Warning;
                case CompilerOutputTab.BestPractices:
                    switch (severity)
                    { 
                        case Severity.Error:
                            return FaultType.BestPracticeError;
                        case Severity.BestPractice:
                            return FaultType.BestPracticeInfo;
                        default:
                            return FaultType.BestPracticeWarning;
                    }
                case CompilerOutputTab.Todo:
                    return FaultType.Todo;
            }

            throw new FileFormatException("Invalid CompilerOutputTab value.");
        }

        /// <summary>
        /// Extracts inner XML from AX compilation output.
        /// </summary>
        /// <param name="html">An HTML string</param>
        /// <returns>An XML string</returns>
        private string ExtractXml(string html)
        {
            // Extract inner XML
            string pattern = @"<AxaptaCompilerOutput>[\w\W]+</AxaptaCompilerOutput>";
            return Regex.Match(html, pattern).Groups[0].Value;
        }

        /// <summary>
        /// Extracts value from the field of the given name.
        /// </summary>
        /// <param name="fields">Collection of fields</param>
        /// <param name="fieldName">Field name</param>
        /// <returns>Field value</returns>
        private string FindValue(IEnumerable<XElement> fields, string fieldName)
        {
            return fields.First(d => d.Attribute("name").Value == fieldName).Value;
        }

        /// <summary>
        /// Loads conetnts of the given file to string.
        /// </summary>
        /// <param name="filename">File to load</param>
        /// <returns>File contents</returns>
        private string GetContents(string filename)
        {
            // Sanitize invalid XML characters right when reading, so we
            // won't have to bother about it later.
            using (FileStream fs = new FileStream(filename, FileMode.Open))
            using (var reader = new XmlSanitizingStream(fs))
            {
                return reader.ReadToEnd();
            }
        }
    }
}
